from PygFW.Event import Event
import pygame

class QuitEvent(Event):

    def __init__(self, surface):

        Event.__init__(self, surface, pygame.QUIT, 'quit_event')

    def executor(self, scene, event):

        scene.run = False


import pygame

pygame.init()

class ImageObject:

    def __repr__(self):

        return f"""<Image Object (path={self.image_path}, dimensions={self.dimensions})>"""

    def __init__(self, image_path, self_proclaimed_id: str=None, convert_to_alpha=True, initial_rotation=None, initial_scale=None, initial_dimensions=None):

        self.image_path = image_path
        self.image_loaded = pygame.image.load(image_path)

        self._self_proclaimed_id = f'<Image Object (path={self.image_path}, alpha={convert_to_alpha}, rotation={initial_rotation})>' if self_proclaimed_id is None else self_proclaimed_id
        self._actual_image_id = self._self_proclaimed_id

        self._image_converted = self.image_loaded.convert_alpha() if convert_to_alpha else self.image_loaded.convert()
        self._initialised_image = self._image_converted

        self.original_dimensions = self.width, self.height
        self.dimensions_history = []

        self.initialised_rotation = initial_rotation
        self.initial_rotation = self.initialised_rotation
        if initial_rotation:

            self.change_rotation(initial_rotation)
            self.initial_rotation = initial_rotation

        self.initialised_scale = 1
        self.initial_scale = self.initialised_scale
        if initial_scale:

            self.change_scale(initial_scale)
            self.initial_scale = initial_scale

        self.initialised_dimensions = (self.width, self.height)
        self.initial_dimensions = self.initialised_dimensions
        if initial_dimensions:

            self.change_dimensions(initial_dimensions)
            self.initial_dimensions = initial_dimensions




    @property
    def user_defined_id(self):

        return self._self_proclaimed_id

    @property
    def actual_image_id(self):

        return self._actual_image_id

    @actual_image_id.setter
    def actual_image_id(self, new_id: str):

        self._actual_image_id = new_id

    @property
    def pygame_rect(self):

        return self._image_converted.get_rect()

    @property
    def width(self):

        return self.pygame_rect.width

    @property
    def height(self):

        return self.pygame_rect.height

    @property
    def dimensions(self):

        return self.width, self.height

    @property
    def image(self):

        return self._image_converted

    def get_warped(self, new_dimensions: [int, int]):

        return pygame.transform.scale(self._initialised_image, new_dimensions)

    def change_dimensions(self, new_dimensions: [int, int]):

        self._image_converted = self.get_warped(new_dimensions)

    def change_scale(self, new_scale: float):

        self._image_converted = self.get_warped((self.width * new_scale, self.height * new_scale))

    def get_rotated(self, new_rotation):

        return pygame.transform.rotate(self._initialised_image, new_rotation)

    def change_rotation(self, new_rotation: int):

        self._image_converted = self.get_rotated(new_rotation)

    def draw(self, pygame_surface: pygame.Surface, spawn: [int, int], offset_x=0, offset_y=0):

        pygame_surface.blit(self._image_converted, (spawn[0] + offset_x, spawn[1] + offset_y))

    def convert(self, convert_to_alpha=True):

        self._image_converted = self.image_loaded.convert_alpha() if convert_to_alpha else self.image_loaded.convert()

    def get_faded(self, new_alpha):

        original_image = self._initialised_image
        faded_image = original_image
        faded_image.set_alpha(new_alpha)

        return faded_image

    def fade(self, new_alpha):

        self._image_converted = self.get_faded(new_alpha)


class Cake:

    def __init__(self):

        self._layers = []

    @property
    def length(self):

        return len(self._layers)

    @property
    def full(self):

        return self._layers

    @property
    def _dict_(self):

        cake_dict = {}

        for index in range(self.length):
            cake_dict[index] = self._layers[index]

        return cake_dict

    def apply(self, obj, *args, index=-1, **kwargs):

        if index < 0:

            index = self.length + 1

        self._layers.insert(index, (obj, args, kwargs))

    def get(self, layer_index, delete=False):

        if layer_index < self.length:

            layer = self._layers[layer_index]

            if delete:

                self._layers.pop(layer_index)

            return layer

        else:

            return None

    def enumerate(self):

        return [(index, self._layers[index]) for index in range(self.length)]

    def destroy(self):

        self._layers = []


_constant_id = 0

def _create_constant_id():

    global _constant_id

    _constant_id += 1

    return f'<Game Constant (id={_constant_id})'

text_relative_margin_size = _create_constant_id()
margin_relative_text_spawn = _create_constant_id()
mutable_text_size = _create_constant_id()


from PygFW.Builtin.Entities.BackgroundEntity import BackgroundEntity
from PygFW.Builtin.Entities.ParticulateEntity import ParticulateEntity
from PygFW.Builtin.Entities.ButtonEntity import ButtonEntity


import pygame
from PygFW import Data, Volatile, Event
from PygFW.Data import DataContainer

class Scene(Volatile.VolatileObject):

    def __init__(self, engine_surface):

        Volatile.VolatileObject.__init__(self)

        self._continue_scene = True

        self.engine_surface = engine_surface

        self.events = DataContainer()
        self.threaded_events = DataContainer()
        self.entities = DataContainer()

        self._coordinate_offset = (0, 0)
        self._coordinate_offset_x = self._coordinate_offset[0]
        self._coordinate_offset_y = self._coordinate_offset[0]

    @property
    def run(self):

        return self._continue_scene

    @run.setter
    def run(self, new_run_status):

        self._continue_scene = new_run_status

    @property
    def coordinate_offset(self):

        return self._coordinate_offset

    @coordinate_offset.setter
    def coordinate_offset(self, new_coordinate_offset):

        new_x, new_y = new_coordinate_offset

        self._coordinate_offset = (new_x, new_y)
        self._coordinate_offset_x = new_x
        self._coordinate_offset_y = new_y

    def map_entities(self):

        entity_coordinates = [entity.relative_position for entity in self.entities._list_]

        mapped_entities = []

        for x in range(self.engine_surface.window_width):

            x_layer = []

            for y in range(self.engine_surface.window_height):

                x_layer.append(0 if (x, y) not in entity_coordinates else 1)

            mapped_entities.append(x_layer)

        return mapped_entities

    def map_entities_in_selection(self, radical_x, a_radical_x, radical_y, a_radical_y):

        entity_coordinates = [entity.relative_position for entity in self.entities._list_]

        mapped_entities = []

        for x in range(radical_x, a_radical_x):

            x_layer = []

            for y in range(radical_y, a_radical_y):

                x_layer.append(0 if (x, y) not in entity_coordinates else 1)

            mapped_entities.append(x_layer)

        return mapped_entities



from PygFW import Event
import pygame

class EntityClickEvent(Event):

    def __init__(self, scene_surface):

        Event.__init__(self, scene_surface, pygame.MOUSEBUTTONDOWN)

    def executor(self, scene, event):

        for entity in scene.entities._list_:

            if entity.collides_with([event.pos]):

                entity.click(scene, event)

class EntityUnclickEvent(Event):

    def __init__(self, scene_surface):

        Event.__init__(self, scene_surface, pygame.MOUSEBUTTONUP)

    def executor(self, scene, event):

        for entity in scene.entities._list_:

            entity.un_click(scene, event)



from PygFW.Data import DataContainer
from PygFW.Data.Manipulator import JoinLists

alphabet = 'abcdefghijklmnopqrstuvwxyz'
alphanumeric = """ 1234567890-=!\"£$%^&*()_+[]#{}~;':@,./<>?|\\"""

class Character:

    def __call__(self, character):

        self.__init__(character)

    def __init__(self, character):

        self._character = character if character in JoinLists(alphabet, alphabet.upper(), alphanumeric) and len(character) == 1 else 'a'

    @property
    def character(self):

        return self._character

    @character.setter
    def character(self, character):

        self.__init__(character)

    @property
    def upper(self):

        return self._character.upper()

    @property
    def lower(self):

        return self._character.lower()

class CharacterCollection:

    def __init__(self, characters=[]):

        self.characters = DataContainer()

        for character in characters:

            self.characters.add(character)

    @property
    def string(self):

        return ''.join(character.character for character in self.characters._list_)

    def construct(self):

        pass

    def strip(self, strip_string, strip_end=False):

        pass




from PygFW.Entity import Entity
from PygFW.Event import Eventifies
from PygFW.Image import ImageGallery, ImageObject
from PygFW.Volatile import default_clock
import pygame

class ButtonEntity(Entity):

    def __init__(self, scene_surface, static_sprite, spawn, highlight_sprite=None, click_sprite=None):

        Entity.__init__(self, scene_surface, ImageGallery(), spawn)

        self.static_sprite = static_sprite
        self.highlight_sprite = highlight_sprite if highlight_sprite else static_sprite
        self.click_sprite = click_sprite if click_sprite else static_sprite

        self.operating_sprite = self.static_sprite

        self.clicked = False

    @property
    def sprite(self):

        return self.operating_sprite

    def draw(self):

        self.operating_sprite.draw(self.scene.engine_surface.pygame_surface, self.relative_position)

    def tick(self):

        self.update()

    def update(self):

        mouse = pygame.mouse.get_pos()

        if self.collides_with([mouse]) and not self.clicked:

            self.operating_sprite = self.highlight_sprite

        elif self.clicked:

            self.operating_sprite = self.click_sprite

        else:

            self.operating_sprite = self.static_sprite

    def click(self, scene, event):

        self.clicked = True
        self.operating_sprite = self.click_sprite

        self.draw()
        pygame.display.update()

    def un_click(self, scene, event):

        self.clicked = False
        self.operating_sprite = self.static_sprite



from PygFW.Entity import Entity
from random import randint
from PygFW import Eventifies

class ParticulateEntity(Entity):

    def __init__(self, scene, sprites, spawn, fall_velocity: [int, int] = (0, 0), wind_velocity: [0, 0] = (0, 0) , justify_existence: bool = True, spawn_radius: [int, int, int, int]=(0, 50, 0, 50), rumble: [int, int]= (-2, 2), spawn_scale: [int, int, int, int] = None, random_scale: [int, int]=(1, 1)):

        Entity.__init__(self, scene, sprites, spawn)

        self.fall_velocity = fall_velocity
        self.wind_velocity = wind_velocity

        self.justify_existence = justify_existence
        self.random_scale = random_scale

        self.a_radical_rumble = rumble[0]
        self.radical_rumble = rumble[1]

        self.a_radical_x = spawn_radius[0]
        self.radical_x = spawn_radius[1]

        self.a_radical_y = spawn_radius[2]
        self.radical_y = spawn_radius[3]

        if spawn_scale:

            self.sprite.change_scale(spawn_scale)

        self.x, self.y = self.calculate_random_spawn()

    def calculate_random_spawn(self):

        return randint(self.a_radical_x, self.radical_x), randint(self.a_radical_y, self.radical_y)

    def rumble(self, a_radical_rumble, radical_rumble):

        self.sprite.change_rotation(randint(a_radical_rumble, radical_rumble))

    def tick(self):

        self.rumble(self.a_radical_rumble, self.radical_rumble)
        self.gravity()
        self.update()

        if self.random_scale:

            self.rescale()

    def gravity(self):

        self.y += randint(*self.fall_velocity)
        self.x += randint(*self.wind_velocity)

    def update(self):

        if self.justify_existence:

            if not 0 < self.y < self.scene.engine_surface.window_height or not 0 < self.x < self.scene.engine_surface.window_width:

                @Eventifies(scene=self.scene, listener=None)
                def remove_self(scene, event):
                    scene.entities.remove(scene.entities.get_tag(self), True)

                self.scene.engine_surface.single_events.add(remove_self)

    def rescale(self):

        self.sprite.change_scale((randint(self.random_scale[0], self.random_scale[1]), randint(self.random_scale[2], self.random_scale[3])))




from PygFW.Image.ImageObject import ImageObject


class ImageBank:

    def __call__(self, image_key):

        return self.get(image_key)

    def __repr__(self):

        return f"""<Image Bank Object (size={self.size}, _dict_={self._dict_})>"""

    def __init__(self):

        self._bank = {}

    @property
    def bank(self):

        return self._bank

    @property
    def size(self):

        return len(self._bank)

    @property
    def _dict_(self):

        self_dict = {}

        index = 0
        for image in self._bank:

            self_dict[index] = image

            index += 1

        return self_dict

    def add(self, image: ImageObject, use_self_proclaimed_id=True, overwrite=True):

        image_id = image.user_defined_id if use_self_proclaimed_id else image.actual_image_id

        if image_id not in self._bank or overwrite:

            self._bank[image_id] = image

            return True

        else:

            return False

    def get(self, image_key) -> ImageObject:

        return self._bank[image_key] if image_key in self._bank else False



from PygFW.Cake import Cake
from PygFW.Scene import Scene
from PygFW.Volatile import VolatileObject, default_clock
from PygFW.Image import ImageBank
from PygFW.Data import DataContainer, Manipulator

import pygame
from _thread import start_new_thread

class Engine(VolatileObject):

    def __init__(self, window_dimensions, fps):

        VolatileObject.__init__(self)

        self._window_dimensions = window_dimensions
        self.pygame_surface = pygame.display.set_mode(window_dimensions)

        self._surface_fps = 1
        self.surface_fps = fps

        self.window_caption = 'A PygFW'
        self.window_icon = None

        self.scenes = DataContainer()
        self._image_bank = ImageBank()

        self.single_events = DataContainer()
        self.repeated_events = DataContainer()

    @property
    def image_bank(self):

        return self._image_bank

    def console_log(self, message, event: str=None):

        event_str = ''

        if event:

            event_str = f'Triggered by event: \'{event}\' '

        print(f""" Console.) {message} {event_str}(time={default_clock})""")

    def console_warn(self, message, event: str=None):

        event_str = ''

        if event:
            event_str = f'Triggered by event: \'{event}\' '

        print(f""" <WARN> Console <WARN>.) {message} {event_str}(time={default_clock})""")

    @property
    def window_dimensions(self):

        return self._window_dimensions

    @window_dimensions.setter
    def window_dimensions(self, new_x_y):

        self._window_dimensions = new_x_y

    @property
    def window_width(self):

        return self._window_dimensions[0]

    @property
    def window_height(self):

        return self._window_dimensions[1]

    @property
    def window_middle_x(self):

        return int(self.window_width / 2)

    @property
    def window_middle_y(self):

        return int(self.window_height / 2)

    @property
    def surface_fps(self):

        return self._surface_fps

    @surface_fps.setter
    def surface_fps(self, fps: int, absolute_set=False):

        if absolute_set:

            self._surface_fps = fps

        else:

            self._surface_fps = int(1000 / fps)

    def start_background_process(self, function, *args, pass_engine=True, **kwargs):

        if pass_engine:

            start_new_thread(function, (self, *args), **kwargs)

        else:

            start_new_thread(function, *args, **kwargs)

    def register_scene(self, scene, scene_id, warn=True):

        if self.scenes.contains_tag(scene_id) and warn:

            self.console_warn(f'Scene with id: \'{scene_id}\' already exists', '*.register_scene()')

        else:

            self.scenes.add(scene, scene_id)

    def get_scene(self, scene_id):

        return self.scenes.get(scene_id) if self.scenes.contains_tag(scene_id) else None

    def start_scene(self, scene_id):

        workspace_scene: Scene = self.get_scene(scene_id)
        workspace_scene.engine_surface = self


        while workspace_scene.run:

            self.pygame_surface = pygame.display.set_mode(self.window_dimensions)
            self.pygame_surface.fill((0, 0, 0))

            pygame.display.set_caption(self.window_caption)
            if self.window_icon:

                pygame.display.set_icon(self.window_icon)

            active_keys = pygame.key.get_pressed()
            pygame_events = pygame.event.get()

            for event in workspace_scene.events._list_:

                for pygame_event in pygame_events:

                    if event.listener == pygame_event.type:

                        event.executor(workspace_scene, pygame_event)

                if active_keys[event.listener]:

                    event.executor(workspace_scene, active_keys)

            for event in workspace_scene.threaded_events._list_:

                for pygame_event in pygame_events:

                    if event.listener == pygame_event.type:

                        start_new_thread(event.executor, (workspace_scene, event))

                if active_keys[event.listener]:

                    start_new_thread(event.executor, (workspace_scene, event))

            for entity in workspace_scene.entities._list_:

                entity.tick()
                entity.draw()

            for event in Manipulator.JoinLists(self.single_events._list_, self.repeated_events._list_):

                event.executor(workspace_scene, event)

            self.single_events.clear()

            pygame.display.update()


def JoinLists(*args: [list]):

    new_list = []

    for arg_list in args:

        if type(arg_list) in [list, tuple, str]:

            for item in arg_list:

                new_list.append(item)

        else:

            raise TypeError(f'{arg_list} is not of type list but is of type \'{type(arg_list)}\'')

    return new_list




from PygFW.Text.Character import Character, CharacterCollection, alphabet, alphanumeric
from PygFW.Text.TextField import TextFieldCharacter, TextFieldCharacterCollection, TextFieldStringConstructor, TextFieldGrid, TextAlertBoxEntity, TextFieldBoxEntity


from PygFW.Text.Character import CharacterCollection, Character
from PygFW import Entity
from PygFW.Data import DataContainer
from PygFW.Constants import *
from PygFW.Event import Eventifies
from PygFW.Volatile import default_clock
import pygame

default_font_path = 'Assets/Fonts/NotoSerif/NotoSerif-Bold.ttf'

def draw_text_absolute(surface, text = 'A', position: [int, int] = (0, 0), colour: [int, int, int] = (255, 255, 255), font_path: str = default_font_path, font_size: int = 18):

    # TODO make a place to store fonts to be reused to reduce lag

    font = pygame.font.Font(font_path, font_size)
    text_display = font.render(text, True, colour)
    # TODO incorporate drawing text into textfield character
    surface.blit(text_display, position)

class TextFieldStringConstructor(DataContainer):

    def __init__(self, string: str, colour: [int, int, int] = (255, 255, 255), font_path: str = default_font_path, font_size: int = 18):

        DataContainer.__init__(self)

        self.colour = colour
        self.font_path = font_path
        self.font_size = font_size

        [self.add(TextFieldCharacter(character, self.colour, self.font_path, self.font_size)) for character in string]

    @property
    def _string_(self):

        string = ''

        for character in self._list_:

            string += character.character

        return string

    @property
    def rgb(self):

        return self.colour

    @property
    def font(self):

        return self.font_path, self.font_size

class TextFieldCharacter(Character):

    def __init__(self, character, colour: [int, int, int] = (255, 255, 255), font_path: str = default_font_path, font_size: int = 18):

        Character.__init__(self, character)
        self.colour = colour
        self.font_path = font_path
        self.font_size = font_size

        self.constructed = self.construct()

    @property
    def rgb(self):

        return self.colour

    @property
    def font(self):

        return self.font_path, self.font_size

    @property
    def rect(self):

        return self.constructed.get_rect()

    def fade(self, fade_amount):

        self.constructed.set_alpha(fade_amount)

    def construct(self):

        # TODO make global font bank
        font = pygame.font.Font(self.font_path, self.font_size)
        return font.render(self.character, True, self.colour)

    def draw(self, surface, position):

        surface.blit(self.constructed, position)


class TextFieldCharacterCollection(CharacterCollection):

    def __init__(self, characters: [TextFieldCharacter, ]):

        CharacterCollection.__init__(self, characters)

    @property
    def width(self):

        combined_width = 0

        for character in self.characters._list_:

            combined_width += character.rect.width

        return combined_width

    @property
    def height(self):

        max_height = 0

        for character in self.characters._list_:

            if (height := character.rect.height) > max_height:

                max_height = height

        return max_height


    def draw(self, surface, position):

        x, y = position

        for character in self.characters._list_:

            character.draw(surface, (x, y))

            x += character.rect.width + 1

    def concatenate(self, character_collection: CharacterCollection):

        print(character_collection)

        for character in character_collection.characters._list_:

            self.characters.add(character)

    def truncate(self, size=48):

        removed_characters = CharacterCollection()

        while self.width > size:

            removed_characters.characters.add(self.characters.remove_index(-1, True))

        return removed_characters if removed_characters.characters.size > 0 else None



class TextFieldBoxEntity(Entity):

    def __init__(self, scene_surface, background_sprites, relative_text_spawn: [int, int] = (5, 5), character_collection: CharacterCollection = CharacterCollection(), spawn: [int, int] = (0, 0)):

        Entity.__init__(self, scene_surface=scene_surface, sprites=background_sprites, spawn=spawn)

        self.string = TextFieldCharacterCollection([])
        self.string.concatenate(character_collection)

        self.relative_text_spawn = relative_text_spawn

    def draw(self):

        #self.sprites.front.draw(self, self.scene.engine_surface.pygame_surface, self.relative_position)

        text_x, text_y = self.relative_position
        rel_x, rel_y = self.relative_text_spawn

        text_x += rel_x
        text_y += rel_y

        self.string.draw(self.scene.engine_surface.pygame_surface, (text_x, text_y))

class TextAlertBoxEntity(TextFieldBoxEntity):

    def __init__(self, scene_surface, background_sprites, relative_text_spawn: [int, int] = (5, 5), character_collection: CharacterCollection = CharacterCollection(), spawn: [int, int] = (0, 0), fade_ferocity: int = 1, fade_interval: int = 1):

        TextFieldBoxEntity.__init__(self, scene_surface=scene_surface, background_sprites=background_sprites, relative_text_spawn=relative_text_spawn, spawn=spawn, character_collection=character_collection)
        self.opacity = 255
        self.fade_ferocity = fade_ferocity
        self.fade_interval = fade_interval

    def update(self):

        if default_clock.milliseconds % self.fade_interval == 0:

            self.opacity -= self.fade_ferocity

        for character in self.string.characters._list_:

            character.fade(self.opacity)

        if self.opacity < 0:

            @Eventifies(scene=self.scene, listener=None)
            def remove_self(scene, event):
                scene.entities.remove(scene.entities.get_tag(self), True)

            self.scene.engine_surface.single_events.add(remove_self)

class TextFieldGrid:

    def __init__(self):

        self._entity_grid = []


import pathlib
from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent

README = (here / 'README.md').read_text()

setup(
    name = 'PygFW',
    author = 'Shaun Cameron',
    author_email = 'shauncameron13034@gmail.com',
    version= '0a',
    description = 'Create a simple yet effective server with and connect to it with a simple client!',
    long_description = README,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/shauncameron/PygFW',
    license = 'MIT',
    packages = find_packages(),
    install_requires = [
        'pygame'
    ]
)

print(' Packages Loaded >>', find_packages())


class DataContainer:

    def __setitem__(self, data_tag, data):

        self.add(data, data_tag)

    def __getitem__(self, data_tag):

        return self.get(data_tag)

    def __call__(self, data_tag):

        return self.get(data_tag)

    def __repr__(self):

        return f"""<Data Container Object (size={self.size}, values={self.container})>"""

    def __init__(self, initial_values: dict=None, size_allowance: int=None):

        self._data_container = {} if initial_values is None else initial_values
        self.size_allowance = -1 if size_allowance is None else size_allowance

        self._cap_sizes = {}

    @property
    def container(self) -> dict:

        return self._data_container

    @property
    def size(self) -> int:

        return len(self._data_container)

    @property
    def _list_(self) -> list:

        return [self._data_container[tag] for tag in self._data_container]

    @property
    def _list_tags_(self) -> list:

        return [tag for tag in self._data_container]

    @property
    def _switch_(self) -> dict:

        switched_dict = {}

        for tag in self._data_container:

            switched_dict[self._data_container[tag]] = tag

        return switched_dict

    @property
    def _wrapped_(self):

        return f'{self.container}'

    @property
    def string(self):

        return ''.join(self._list_) if [True*self.size] == [isinstance(item, str) for item in self._list_] else None

    def add(self, data, data_tag: str=None, overwrite=True):

        data_tag = f'<DataContainerItem (index={self.size}, data={data})>' if data_tag is None else data_tag

        if data_tag is None:

            raise ValueError(f'Data tag belonging to :\'{data}\' is None, this means no data_tag was given in \'MDTContainer.add)\' and no data_tag value was given to the MDTObject')

        if data_tag not in self._data_container or overwrite:

            if self.size_allowance > self.size or self.size_allowance < 0:


                if self.capped(data.__class__) and len(self.findall(data.__class__)) < self._cap_sizes[data.__class__]:

                    self._data_container[data_tag] = data

                    return data_tag

                elif not self.capped(data.__class__):

                    self._data_container[data_tag] = data

                    return data_tag

                else:

                    return None

            else:

                return None

        else:

            return None

    def contains(self, data) -> bool:

        return data in self._list_

    def contains_tag(self, data_tag) -> bool:

        return data_tag in self._list_tags_

    def get(self, data_tag):

        return self._data_container[data_tag] if data_tag in self._data_container else None

    def get_tag(self, data) -> str:

        return self._switch_[data]

    def clear(self):

        self._data_container = {}

    def remove(self, data_tag, i_know_what_im_doing=False):

        if data_tag in self._data_container and i_know_what_im_doing:

            data = self._data_container[data_tag]

            del self._data_container[data_tag]

            return data

    def remove_index(self, index, i_know_what_im_doing=False):

        if index < self.size and i_know_what_im_doing:

            data = self._data_container[self._list_tags_[index]]

            del self._data_container[self._list_tags_[index]]

            return data

    def cap(self, object_class, amount):

        self._cap_sizes[object_class] = amount

    def capped(self, object_class):

        return self._cap_sizes[object_class] if object_class in self._cap_sizes else None

    def findall(self, object_type):

        found_objects = []

        for found_object in self._list_:

            if found_object.__class__ == object_type:

                found_objects.append(found_object)

        return found_objects



from PygFW.Image import ImageGallery, ImageObject
from PygFW.Volatile import VolatileObject
from PygFW.Event import Eventifies


class Entity():

    def __init__(self, scene_surface, sprites: [ImageObject, ], spawn: [int, int], rotate_sprites=False,
                 coordinate_offset: [int, int] = (0, 0)):

        # VolatileObject.__init__(self)

        self.scene = scene_surface

        self.sprites = ImageGallery()

        if sprites:

            for sprite in sprites:

                self.sprites.add(sprite)

        self.rotate_sprites = rotate_sprites
        self.sprite_index = 0

        self.x = spawn[0]
        self.y = spawn[1]
        self.x_offset = coordinate_offset[0]
        self.y_offset = coordinate_offset[1]

        self._radius_history_x = self.x
        self._radius_history_y = self.y
        self._radius_history = []

    @property
    def absolute_position(self):

        return self.x, self.y

    @property
    def relative_x(self):

        return self.x + self.x_offset

    @property
    def relative_y(self):

        return self.y + self.y_offset

    @property
    def relative_position(self):

        return self.relative_x, self.relative_y

    @property
    def sprite(self):

        return self.sprites[self.sprite_index]

    @property
    def rect(self):

        return self.sprite.pygame_rect

    @property
    def radius(self):

        appearing_coords = []

        if self.x != self._radius_history_x or self.y != self._radius_history_y or len(self._radius_history) < 1:

            for x in range(int(self.x), int(self.x + self.sprite.width)):

                for y in range(int(self.y), int(self.y + self.sprite.height)):
                    appearing_coords.append((x, y))

        else:

            appearing_coords = self._radius_history

        return appearing_coords

    def collides_with(self, other_radius_list):

        collides = False

        for xy in other_radius_list:

            if xy in self.radius:

                collides = True

        return collides

    def draw(self):

        self.sprites[self.sprite_index].draw(self.scene.engine_surface.pygame_surface, self.relative_position)

        if self.rotate_sprites:
            self.sprites.rotate(1)

    def tick(self):

        self.update()
        self.gravity()

    def update(self):

        pass

    def gravity(self):

        pass

    def hover(self, scene, event):

        pass

    def un_hover(self, scene, event):

        pass

    def click(self, scene, event):

        pass

    def left_click(self, scene, event):

        pass

    def right_click(self, scene, event):

        pass

    def un_click(self, scene, event):

        pass

    def un_left_click(self, scene, event):

        pass

    def un_right_click(self, scene, event):

        pass

    def go_towards(self, coordinates, drift_velocity=1, step_velocity=1, collision_check=False):

        target_x, target_y = coordinates

        if (collision_check and True not in [self.collides_with(other_entity.radius) for other_entity in
                                            self.scene.entities._list_]) or not collision_check:

            if target_x < self.x:

                self.x -= step_velocity

            elif target_x > self.x:

                self.x += step_velocity

            if target_y < self.y:

                self.y -= drift_velocity

            elif target_y > self.y:

                self.y += drift_velocity


from PygFW.Volatile import VolatileObject
import pygame

class Event(VolatileObject):

    def __call__(self, scene, event):

        self.executor(scene, event)

    def __init__(self, scene_surface, listener, event_id: str = None):

        self.listener = listener
        self.surface = scene_surface
        self.event_id = f'<Event Object (listener={listener}, scene={scene_surface}' if event_id is None else event_id

        self.executor_maps = {}

    def executor(self, scene, event):

        for mapped_function in self.executor_maps:

            mapped_function['function'](scene, event, *mapped_function['args'], *mapped_function['kwargs'])

    def maps(self, *args, **kwargs):

        def add_function(function):

            self.executor_maps[function] = {'function': function, 'args': args, 'kwargs': kwargs}

            return function

        return add_function


def Eventifies(scene, listener=None, add=False):

    def set_executor(function):

        event = Event(scene_surface=scene, listener=listener)
        event.executor = function

        return event

    if add:

        scene.events.add(set_executor)

    return set_executor


from PygFW.Entity import Entity
from PygFW.Image import ImageGallery, ImageObject
from PygFW.Volatile import default_clock

class BackgroundEntity(Entity):

    def __init__(self, game_surface, background_sprites: ImageGallery, scroll: bool = True, scroll_velocity: int = -1, millisecond_scroll_interval: int=1):

        Entity.__init__(self, game_surface, background_sprites, (0, 0))

        self.scroll = scroll
        self.scroll_velocity = scroll_velocity
        self.millisecond_scroll_interval = millisecond_scroll_interval

    def draw(self):

        if self.scroll:

            if default_clock.fmilliseconds % self.millisecond_scroll_interval == 0:

                self.x_offset += self.scroll_velocity

            if self.scroll_velocity < 0:

                if self.x_offset + self.sprite.width < self.sprite.width/2:

                    self.x_offset = 0

            elif self.scroll_velocity > 0:

                if self.x_offset - self.sprite.width / 2 == 0:

                    self.x_offset = 0

            self.sprite.draw(self.scene.engine_surface.pygame_surface, self.relative_position, self.x_offset)
            self.sprite.draw(self.scene.engine_surface.pygame_surface, self.relative_position, self.x_offset + self.sprite.width)
            self.sprite.draw(self.scene.engine_surface.pygame_surface, self.relative_position, self.x_offset - self.sprite.width)

        else:

            Entity.draw(self)


from PygFW.Data import DataContainer
from _thread import start_new_thread
from time import sleep


class Clock:

    def __init__(self, run_clock=True, initial_clock_ticks=1):

        self._clock = initial_clock_ticks

        self._run_clock = run_clock
        start_new_thread(self._start_, ())

    @property
    def fmilliseconds(self):

        return self._clock

    @property
    def milliseconds(self):
        return int(self.fmilliseconds)

    @property
    def fseconds(self):

        return self.fmilliseconds / 1000

    @property
    def seconds(self):

        return int(self.fseconds)

    @property
    def fminutes(self):

        return self.fseconds / 60

    @property
    def minutes(self):

        return int(self.fminutes)

    @property
    def fhours(self):

        return self.fminutes / 60

    @property
    def hours(self):

        return int(self.fhours)

    def tick(self, amount=1):

        self._clock += amount

    def _start_(self):

        while self._run_clock:

            sleep(.001)

            self.tick(1)

    def _stop_(self):

        self._run_clock = False

    def create_stamp(self):

        return f'<Clock Time Stamp (milliseconds={self.fmilliseconds}, seconds={self.fseconds}, minutes={self.fminutes}, hours={self.fhours})>', self.fmilliseconds, self.fseconds, self.fminutes, self.fhours


default_clock = Clock()


class VolatileObject:

    def __getitem__(self, data_tag):

        return self.mdt[data_tag]

    def __init__(self):

        self._mdt_container = DataContainer()
        self._object_uptime_clock = Clock()
        self._creation_time = default_clock.create_stamp()

    @property
    def mdt(self):

        return self._mdt_container

    @property
    def uptime(self):

        return self._object_uptime_clock

    @property
    def created_at(self):

        return self._creation_time



from PygFW.Cake import Cake
from PygFW.Engine import Engine
from PygFW.Scene import Scene
from PygFW.Event import Event, Eventifies
from PygFW.Volatile import default_clock, Clock, VolatileObject
from PygFW.Entity import Entity
from PygFW.Constants import *
import PygFW.Image
import PygFW.Data
import PygFW.Text
import PygFW.Builtin


import pygame
from PygFW.Image import ImageObject, ImageBank
from collections import deque


class ImageGallery:

    def __setitem__(self, index, image_object):

        self.add(image_object, index)

    def __getitem__(self, index: int):

        return self._image_gallery[index]

    def __call__(self, index: int):

        return self._image_gallery[index]

    def __init__(self, initial_images: [ImageObject, ]=[]):

        self._image_gallery = deque(initial_images)

    @property
    def size(self):

        return len(self._image_gallery)

    @property
    def front(self):

        return self._image_gallery[0] if self.size > 1 else None

    @property
    def back(self):

        return self._image_gallery[self.size - 1] if self.size > 1 else None

    @property
    def gallery(self):

        return self._image_gallery

    def add(self, image_object: ImageObject, index=-1):

        if index < 0:

            index = self.size + 1

        self._image_gallery.insert(index, image_object)

    def rotate(self, direction):

        if direction > 0:

            self._image_gallery.rotate(1)

        else:

            self._image_gallery.rotate(-1)
